@@ -407,7 +407,7 @@ class QuadraticModelBase {
             en += u_val * linear(u);
 
             auto span = neighborhood(u);
-            for (auto nit = span.first; nit != span.second && (*nit).first < u;
+            for (auto nit = span.first; nit != span.second && (*nit).first <= u;
                  ++nit) {
                 auto v_val = *(sample_start + (*nit).first);
                 auto bias = (*nit).second;
@@ -1006,6 +1006,10 @@ class BinaryQuadraticModel : public QuadraticModelBase<Bias, Index> {
         vartype_ = vartype;
     }
 
+    bias_type lower_bound(index_type v) const {
+        return vartype_info<bias_type>::default_min(this->vartype_);
+    }
+
     /**
      *  Return the number of interactions in the quadratic model.
      * 
@@ -1039,6 +1043,10 @@ class BinaryQuadraticModel : public QuadraticModelBase<Bias, Index> {
         }
     }
 
+    bias_type upper_bound(index_type v) const {
+        return vartype_info<bias_type>::default_max(this->vartype_);
+    }
+
     /// Return the vartype of the binary quadratic model.
     const Vartype& vartype() const { return vartype_; }
 
