@@ -12,10 +12,13 @@
 #    See the License for the specific language governing permissions and
 #    limitations under the License.
 
+from __future__ import annotations
+
 import struct
 import tempfile
+import typing
 
-from collections.abc import Callable
+from collections.abc import Callable, Set
 from copy import deepcopy
 from numbers import Number
 from typing import Any, Dict, Iterator, Iterable, Mapping, Optional, Sequence, Sized, Tuple, Union
@@ -43,7 +46,7 @@ from dimod.views.quadratic import QuadraticViewsMixin
 
 if TYPE_CHECKING:
     # avoid circular imports
-    from dimod import BinaryQuadraticModel
+    from dimod import BinaryQuadraticModel, ConstrainedQuadraticModel
 
 
 __all__ = ['QuadraticModel', 'QM', 'Integer', 'Integers', 'IntegerArray']
@@ -421,7 +424,7 @@ class QuadraticModel(QuadraticViewsMixin):
         Raises:
             ValueError: If a specified variable is not in the model.
         """
-        add_linear = self.add_linear
+        add_linear = self.data.add_linear
 
         if isinstance(linear, Mapping):
             for v, bias in linear.items():
@@ -521,8 +524,56 @@ class QuadraticModel(QuadraticViewsMixin):
 
         """
         vartype = as_vartype(vartype, extended=True)
+        add_variable = self.data.add_variable
         for v in variables:
-            self.add_variable(vartype, v)
+            add_variable(vartype, v)
+
+    def add_variables_from_model(self,
+                                 model: Union[BinaryQuadraticModel,
+                                              ConstrainedQuadraticModel,
+                                              QuadraticModel],
+                                 *,
+                                 variables: Optional[Iterable[Variable]] = None,
+                                 ):
+        """Add variables from another model.
+
+        Args:
+            model: A binary quadratic model, constrained quadratic model or
+                quadratic model.
+
+            variables: The variables from the model to add. If not specified
+                all of the variables are added.
+
+        Examples:
+            >>> qm0 = dimod.Integer('i', lower_bound=5, upper_bound=10) + dimod.Binary('x')
+            >>> qm1 = dimod.QuadraticModel()
+            >>> qm1.add_variables_from_model(qm0)
+            >>> qm1.variables
+            Variables(['i', 'x'])
+            >>> qm1.lower_bound('i')
+            5.0
+
+        """
+        # avoid circular import
+        from dimod.binary import BinaryQuadraticModel
+        from dimod.constrained import ConstrainedQuadraticModel
+
+        if isinstance(model, ConstrainedQuadraticModel):
+            return self.add_variables_from_model(model.objective, variables=variables)
+
+        if variables is None:
+            variables = model.variables
+
+        vartype = model.vartype if isinstance(model, QuadraticModel) else lambda v: model.vartype
+
+        for v in variables:
+            vt = vartype(v)
+            if vt is Vartype.SPIN or vt is Vartype.BINARY:
+                self.add_variable(vt, v)
+            else:
+                self.add_variable(vt, v,
+                                  lower_bound=model.lower_bound(v),
+                                  upper_bound=model.upper_bound(v))
 
     def change_vartype(self, vartype: VartypeLike, v: Variable) -> "QuadraticModel":
         """Change the variable type of the given variable, updating the biases.
@@ -1266,7 +1317,7 @@ class QuadraticModel(QuadraticViewsMixin):
         file.seek(0)
         return file
 
-    def update(self, other: 'QuadraticModel'):
+    def update(self, other: typing.Union[QuadraticModel, BinaryQuadraticModel]):
         """Update the quadratic model from another quadratic model.
 
         Adds to the quadratic model the variables, linear biases, quadratic biases,
@@ -1296,23 +1347,53 @@ class QuadraticModel(QuadraticViewsMixin):
             >>> print(qm1.get_quadratic('s2', 's1'), qm1.get_quadratic('s3', 's1'))
             -2.0 1.0
         """
-        # this can be improved a great deal with c++, but for now let's use
-        # python for simplicity
+        try:
+            return self.data.update(other.data)
+        except (AttributeError, TypeError):
+            pass
+
+        # looks like we have a model that either has object dtype or isn't
+        # a cython model we recognize, so let's fall back on python
+
+        # need a couple methods to be generic between bqm and qm
+        vartype = other.vartype if callable(other.vartype) else lambda v: other.vartype
+
+        def lower_bound(v: Variable) -> Bias:
+            try:
+                return other.lower_bound(v)
+            except AttributeError:
+                pass
+
+            if other.vartype is Vartype.SPIN:
+                return -1
+            elif other.vartype is Vartype.BINARY:
+                return 0
+            else:
+                raise RuntimeError  # shouldn't ever happen
+
+        def upper_bound(v: Variable) -> Bias:
+            try:
+                return other.upper_bound(v)
+            except AttributeError:
+                pass
+
+            return 1
 
         for v in other.variables:
             if v not in self.variables:
                 continue
-            if self.vartype(v) != other.vartype(v):
+
+            if self.vartype(v) != vartype(v):
                 raise ValueError(f"conflicting vartypes: {v!r}")
-            if self.lower_bound(v) != other.lower_bound(v):
+            if self.lower_bound(v) != lower_bound(v):
                 raise ValueError(f"conflicting lower bounds: {v!r}")
-            if self.upper_bound(v) != other.upper_bound(v):
+            if self.upper_bound(v) != upper_bound(v):
                 raise ValueError(f"conflicting upper bounds: {v!r}")
 
         for v in other.variables:
-            self.add_linear(self.add_variable(other.vartype(v), v,
-                                              lower_bound=other.lower_bound(v),
-                                              upper_bound=other.upper_bound(v)),
+            self.add_linear(self.add_variable(vartype(v), v,
+                                              lower_bound=lower_bound(v),
+                                              upper_bound=upper_bound(v)),
                             other.get_linear(v))
 
         for u, v, bias in other.iter_quadratic():
